@@ -1,5 +1,7 @@
+import django
 from django.db import connections
-from django.db.models import Value
+from django.db.models import Expression, F, QuerySet, Value, Window
+from django.db.models.functions import RowNumber
 from django.db.models.sql.compiler import SQLCompiler
 from django.db.models.sql.query import Query
 
@@ -12,8 +14,33 @@ def _find_tree_model(cls):
 
 
 class TreeQuery(Query):
-    # Set by TreeQuerySet.order_siblings_by
-    sibling_order = None
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._setup_query()
+
+    def _setup_query(self):
+        """
+        Run on initialization and at the end of chaining. Any attributes that
+        would normally be set in __init__() should go here instead.
+        """
+        # We add the variables for `sibling_order` and `rank_table_query` here so they
+        # act as instance variables which do not persist between user queries
+        # the way class variables do
+
+        # Only add the sibling_order attribute if the query doesn't already have one to preserve cloning behavior
+        if not hasattr(self, "sibling_order"):
+            # Add an attribute to control the ordering of siblings within trees
+            opts = _find_tree_model(self.model)._meta
+            self.sibling_order = opts.ordering if opts.ordering else opts.pk.attname
+
+        # Only add the rank_table_query attribute if the query doesn't already have one to preserve cloning behavior
+        if not hasattr(self, "rank_table_query"):
+            # Create a default QuerySet for the rank_table to use
+            # so we can avoid recursion
+            self.rank_table_query = QuerySet(model=_find_tree_model(self.model))
+
+        if not hasattr(self, "tree_fields"):
+            self.tree_fields = {}
 
     def get_compiler(self, using=None, connection=None, **kwargs):
         # Copied from django/db/models/sql/query.py
@@ -28,37 +55,37 @@ class TreeQuery(Query):
         return TreeCompiler(self, connection, using, **kwargs)
 
     def get_sibling_order(self):
-        if self.sibling_order is not None:
-            return self.sibling_order
-        opts = _find_tree_model(self.model)._meta
-        if opts.ordering:
-            return opts.ordering
-        return opts.pk.attname
+        return self.sibling_order
+
+    def get_rank_table_query(self):
+        return self.rank_table_query
+
+    def get_tree_fields(self):
+        return self.tree_fields
 
 
 class TreeCompiler(SQLCompiler):
     CTE_POSTGRESQL = """
     WITH RECURSIVE __rank_table(
+        {tree_fields_columns}
         "{pk}",
         "{parent}",
         "rank_order"
     ) AS (
-        SELECT
-            {rank_pk},
-            {rank_parent},
-            ROW_NUMBER() OVER (ORDER BY {rank_order_by})
-        FROM {rank_from}
+        {rank_table}
     ),
     __tree (
+        {tree_fields_names}
         "tree_depth",
         "tree_path",
         "tree_ordering",
         "tree_pk"
     ) AS (
         SELECT
-            0 AS tree_depth,
-            array[T.{pk}] AS tree_path,
-            array[T.rank_order] AS tree_ordering,
+            {tree_fields_initial}
+            0,
+            array[T.{pk}],
+            array[T.rank_order],
             T."{pk}"
         FROM __rank_table T
         WHERE T."{parent}" IS NULL
@@ -66,7 +93,8 @@ class TreeCompiler(SQLCompiler):
         UNION ALL
 
         SELECT
-            __tree.tree_depth + 1 AS tree_depth,
+            {tree_fields_recursive}
+            __tree.tree_depth + 1,
             __tree.tree_path || T.{pk},
             __tree.tree_ordering || T.rank_order,
             T."{pk}"
@@ -76,15 +104,23 @@ class TreeCompiler(SQLCompiler):
     """
 
     CTE_MYSQL = """
-    WITH RECURSIVE __rank_table({pk}, {parent}, rank_order) AS (
-        SELECT
-            {rank_pk},
-            {rank_parent},
-            ROW_NUMBER() OVER (ORDER BY {rank_order_by})
-        FROM {rank_from}
+    WITH RECURSIVE __rank_table(
+        {tree_fields_columns}
+        {pk},
+        {parent},
+        rank_order
+    ) AS (
+        {rank_table}
     ),
-    __tree(tree_depth, tree_path, tree_ordering, tree_pk) AS (
+    __tree(
+        {tree_fields_names}
+        tree_depth,
+        tree_path,
+        tree_ordering,
+        tree_pk
+    ) AS (
         SELECT
+            {tree_fields_initial}
             0,
             -- Limit to max. 50 levels...
             CAST(CONCAT("{sep}", {pk}, "{sep}") AS char(1000)),
@@ -97,6 +133,7 @@ class TreeCompiler(SQLCompiler):
         UNION ALL
 
         SELECT
+            {tree_fields_recursive}
             __tree.tree_depth + 1,
             CONCAT(__tree.tree_path, T2.{pk}, "{sep}"),
             CONCAT(__tree.tree_ordering, LPAD(CONCAT(T2.rank_order, "{sep}"), 20, "0")),
@@ -106,19 +143,27 @@ class TreeCompiler(SQLCompiler):
     )
     """
 
-    CTE_SQLITE3 = """
-    WITH RECURSIVE __rank_table({pk}, {parent}, rank_order) AS (
-        SELECT
-            {rank_pk},
-            {rank_parent},
-            row_number() OVER (ORDER BY {rank_order_by})
-        FROM {rank_from}
+    CTE_SQLITE = """
+    WITH RECURSIVE __rank_table(
+        {tree_fields_columns}
+        {pk},
+        {parent},
+        rank_order
+    ) AS (
+        {rank_table}
     ),
-    __tree(tree_depth, tree_path, tree_ordering, tree_pk) AS (
+    __tree(
+        {tree_fields_names}
+        tree_depth,
+        tree_path,
+        tree_ordering,
+        tree_pk
+    ) AS (
         SELECT
-            0 tree_depth,
-            printf("{sep}%%s{sep}", {pk}) tree_path,
-            printf("{sep}%%020s{sep}", T.rank_order) tree_ordering,
+            {tree_fields_initial}
+            0,
+            printf("{sep}%%s{sep}", {pk}),
+            printf("{sep}%%020s{sep}", T.rank_order),
             T."{pk}" tree_pk
         FROM __rank_table T
         WHERE T."{parent}" IS NULL
@@ -126,6 +171,7 @@ class TreeCompiler(SQLCompiler):
         UNION ALL
 
         SELECT
+            {tree_fields_recursive}
             __tree.tree_depth + 1,
             __tree.tree_path || printf("%%s{sep}", T.{pk}),
             __tree.tree_ordering || printf("%%020s{sep}", T.rank_order),
@@ -135,13 +181,8 @@ class TreeCompiler(SQLCompiler):
     )
     """
 
-    def get_sibling_order_params(self):
-        """
-        This method uses a simple django queryset to generate sql
-        that can be used to create the __rank_table that orders
-        siblings. This is done so that any joins required by order_by
-        are pre-calculated by django
-        """
+    def get_rank_table(self):
+        # Get and validate sibling_order
         sibling_order = self.query.get_sibling_order()
 
         if isinstance(sibling_order, (list, tuple)):
@@ -153,38 +194,41 @@ class TreeCompiler(SQLCompiler):
                 "Sibling order must be a string or a list or tuple of strings."
             )
 
-        # Use Django to make a SQL query whose parts can be repurposed for __rank_table
-        base_query = (
-            _find_tree_model(self.query.model)
-            .objects.only("pk", "parent")
-            .order_by(*order_fields)
-            .query
+        # Convert strings to expressions. This is to maintain backwards compatibility
+        # with Django versions < 4.1
+        if django.VERSION < (4, 1):
+            base_order = []
+            for field in order_fields:
+                if isinstance(field, Expression):
+                    base_order.append(field)
+                elif isinstance(field, str):
+                    if field[0] == "-":
+                        base_order.append(F(field[1:]).desc())
+                    else:
+                        base_order.append(F(field).asc())
+            order_fields = base_order
+
+        # Get the rank table query
+        rank_table_query = self.query.get_rank_table_query()
+
+        rank_table_query = (
+            rank_table_query.order_by()  # Ensure there is no ORDER BY at the end of the SQL
+            # Values allows us to both limit and specify the order of
+            # the columns selected so that they match the CTE
+            .values(
+                *self.query.get_tree_fields().values(),
+                "pk",
+                "parent",
+                rank_order=Window(
+                    expression=RowNumber(),
+                    order_by=order_fields,
+                ),
+            )
         )
 
-        # Use the base compiler because we want vanilla sql and want to avoid recursion.
-        base_compiler = SQLCompiler(base_query, self.connection, None)
-        base_sql, base_params = base_compiler.as_sql()
-        result_sql = base_sql % base_params
+        rank_table_sql, rank_table_params = rank_table_query.query.sql_with_params()
 
-        # Split the base SQL string on the SQL keywords 'FROM' and 'ORDER BY'
-        from_split = result_sql.split("FROM")
-        order_split = from_split[1].split("ORDER BY")
-
-        # Identify the FROM and ORDER BY parts of the base SQL
-        ordering_params = {
-            "rank_from": order_split[0].strip(),
-            "rank_order_by": order_split[1].strip(),
-        }
-
-        # Identify the primary key field and parent_id field from the SELECT section
-        base_select = from_split[0][6:]
-        for field in base_select.split(","):
-            if "parent_id" in field:  # XXX Taking advantage of Hardcoded.
-                ordering_params["rank_parent"] = field.strip()
-            else:
-                ordering_params["rank_pk"] = field.strip()
-
-        return ordering_params
+        return rank_table_sql, rank_table_params
 
     def as_sql(self, *args, **kwargs):
         # Try detecting if we're used in a EXISTS(1 as "a") subquery like
@@ -229,8 +273,39 @@ class TreeCompiler(SQLCompiler):
             "sep": SEPARATOR,
         }
 
-        # Add ordering params to params
-        params.update(self.get_sibling_order_params())
+        # Get the rank_table SQL and params
+        rank_table_sql, rank_table_params = self.get_rank_table()
+        params["rank_table"] = rank_table_sql
+
+        if self.connection.vendor == "postgresql":
+            cte = self.CTE_POSTGRESQL
+            cte_initial = "array[T.{column}]::text[], "
+            cte_recursive = "__tree.{name} || T.{column}::text, "
+        elif self.connection.vendor == "sqlite":
+            cte = self.CTE_SQLITE
+            cte_initial = 'printf("{sep}%%s{sep}", {column}), '
+            cte_recursive = '__tree.{name} || printf("%%s{sep}", T.{column}), '
+        elif self.connection.vendor == "mysql":
+            cte = self.CTE_MYSQL
+            cte_initial = 'CAST(CONCAT("{sep}", {column}, "{sep}") AS char(1000)), '
+            cte_recursive = 'CONCAT(__tree.{name}, T2.{column}, "{sep}"), '
+
+        tree_fields = self.query.get_tree_fields()
+        qn = self.connection.ops.quote_name
+        params.update({
+            "tree_fields_columns": "".join(
+                f"{qn(column)}, " for column in tree_fields.values()
+            ),
+            "tree_fields_names": "".join(f"{qn(name)}, " for name in tree_fields),
+            "tree_fields_initial": "".join(
+                cte_initial.format(column=qn(column), name=qn(name), sep=SEPARATOR)
+                for name, column in tree_fields.items()
+            ),
+            "tree_fields_recursive": "".join(
+                cte_recursive.format(column=qn(column), name=qn(name), sep=SEPARATOR)
+                for name, column in tree_fields.items()
+            ),
+        })
 
         if "__tree" not in self.query.extra_tables:  # pragma: no branch - unlikely
             tree_params = params.copy()
@@ -246,16 +321,16 @@ class TreeCompiler(SQLCompiler):
             if aliases:
                 tree_params["db_table"] = aliases[0]
 
+            select = {
+                "tree_depth": "__tree.tree_depth",
+                "tree_path": "__tree.tree_path",
+                "tree_ordering": "__tree.tree_ordering",
+            }
+            select.update({name: f"__tree.{name}" for name in tree_fields})
             self.query.add_extra(
                 # Do not add extra fields to the select statement when it is a
                 # summary query or when using .values() or .values_list()
-                select={}
-                if skip_tree_fields or self.query.values_select
-                else {
-                    "tree_depth": "__tree.tree_depth",
-                    "tree_path": "__tree.tree_path",
-                    "tree_ordering": "__tree.tree_ordering",
-                },
+                select={} if skip_tree_fields or self.query.values_select else select,
                 select_params=None,
                 where=["__tree.tree_pk = {db_table}.{pk}".format(**tree_params)],
                 params=None,
@@ -269,27 +344,29 @@ class TreeCompiler(SQLCompiler):
                 ),
             )
 
-        if self.connection.vendor == "postgresql":
-            cte = self.CTE_POSTGRESQL
-        elif self.connection.vendor == "sqlite":
-            cte = self.CTE_SQLITE3
-        elif self.connection.vendor == "mysql":
-            cte = self.CTE_MYSQL
         sql_0, sql_1 = super().as_sql(*args, **kwargs)
         explain = ""
         if sql_0.startswith("EXPLAIN "):
             explain, sql_0 = sql_0.split(" ", 1)
 
-        return ("".join([explain, cte.format(**params), sql_0]), sql_1)
+        # Pass any additional rank table sql paramaters so that the db backend can handle them.
+        # This only works because we know that the CTE is at the start of the query.
+        return (
+            "".join([explain, cte.format(**params), sql_0]),
+            rank_table_params + sql_1,
+        )
 
     def get_converters(self, expressions):
         converters = super().get_converters(expressions)
+        tree_fields = {"__tree.tree_path", "__tree.tree_ordering"} | {
+            f"__tree.{name}" for name in self.query.tree_fields
+        }
         for i, expression in enumerate(expressions):
             # We care about tree fields and annotations only
             if not hasattr(expression, "sql"):
                 continue
 
-            if expression.sql in {"__tree.tree_path", "__tree.tree_ordering"}:
+            if expression.sql in tree_fields:
                 converters[i] = ([converter], expression)
         return converters
 
